#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#ifdef _MSC_VER
#if _MSC_VER >= 1200
#pragma comment(linker, "/OPT:NOWIN98")
#endif
#endif

static unsigned char buf[4096];
typedef struct PATCH_TAG {
	int addr;
	int data;
	struct PATCH_TAG *next;
} PATCH;

PATCH *patch_top = 0;
static void free_patch(void) {
	PATCH *cur, *next;
	for (cur = patch_top; cur; cur = next)
	{
		next = cur->next;
		free(cur);
	}
	patch_top = 0;
}

static void alloc_patch(int addr, int data)
{
	PATCH *cur = malloc(sizeof(PATCH));
	if (!cur) return;
	cur->addr = addr;
	cur->data = data;
	cur->next = patch_top;
	patch_top = cur;
}

static void patch_patch(int base, unsigned char *p, int size)
{
	int limit = base + size;
	PATCH *cur;
	for (cur = patch_top; cur; cur = cur->next)
	{
		if (base <= cur->addr && cur->addr < limit)
		{
			p[cur->addr - base] = cur->data;
		}
	}
}

static char *hatoi2(char *p, int *pi)
{
	int c, r;
	while (*p == ' ' || *p == '\t') p++;
	for (r = 0; *p; p++)
	{
		c = *p;
		switch (c)
		{
			case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
				c = c + 'A' - 'a';
			case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
				c = c + '0' + 10 - 'A';
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				r = r * 16 + (c - '0');
				continue;
		}
		break;
	}
	*pi = r;
	return p;
}

static int hatoi(char *p)
{
	int ret;
	hatoi2(p, &ret);
	return ret;
}

int main(int argc, char **argv)
{
	int i, skip = 0, len = 0, unit = 0, num = 0, block = 0, help = 1, zero = 0;
	char *outpath = NULL;
	for (i = 1; i < argc; i++)
	{
		char *p = argv[i];
		if (strncmp(p, "-o", 2) == 0)
		{
			outpath = p[2] ? p + 2 : argv[++i];
		}
		else if (strncmp(p, "-s", 2) == 0)
		{
			p = p[2] ? p + 2 : argv[++i];
			skip = hatoi(p);
		}
		else if (strncmp(p, "-l", 2) == 0)
		{
			p = p[2] ? p + 2 : argv[++i];
			len = hatoi(p);
		}
		else if (strncmp(p, "-u", 2) == 0)
		{
			p = p[2] ? p + 2 : argv[++i];
			unit = hatoi(p);
		}
		else if (strncmp(p, "-n", 2) == 0)
		{
			p = p[2] ? p + 2 : argv[++i];
			num = hatoi(p);
		}
		else if (strncmp(p, "-b", 2) == 0)
		{
			p = p[2] ? p + 2 : argv[++i];
			block = hatoi(p);
		}
		else if (strncmp(p, "-p", 2) == 0)
		{
			int addr, data;
			p = p[2] ? p + 2 : argv[++i];
			p = hatoi2(p, &addr);
			p = (*p == ':') ? p + 1 : argv[++i];
			do {
				p = hatoi2(p, &data);
				alloc_patch(addr++, data);
			} while (*p++ == ',');
		}
		else if (strncmp(p, "-z", 2) == 0)
		{
			p = p[2] ? p + 2 : argv[++i];
			zero = hatoi(p);
		}
		else
		{
			FILE *ifp, *ofp;
			int base = 0, size, uskip = 0;
			ifp = fopen(p,"rb");
			if(!ifp) continue;
			fseek(ifp, 0, SEEK_END);
			size = ftell(ifp) - skip;
			if (unit)
			{
				uskip = unit * block + (size % unit);
				size -= uskip;
			}
			if (size  < 1)
			{
				fclose(ifp);
				continue;
			}
			if (outpath)
			{
				ofp = fopen(outpath,"wb");
				if (!ofp)
				{
					fclose(ifp);
					continue;
				}
			}
			else
			{
				ofp = stdout;
			}
			fseek(ifp, skip + uskip, SEEK_SET);
			if (len && size > len) size = len;
			if (num * unit && size > num * unit) size = num * unit;
			while (size > sizeof(buf))
			{
				fread(buf, 1, sizeof(buf), ifp);
				patch_patch(base, buf, sizeof(buf));
				fwrite(buf, 1, sizeof(buf), ofp);
				size -= sizeof(buf);
				base += sizeof(buf);
			}
			fread(buf, 1, size, ifp);
			patch_patch(base, buf, size);
			fwrite(buf, 1, size, ofp);
			if (zero)
			{
				memset(buf, 0, sizeof(buf));
				for (size = zero; size > sizeof(buf); size -= sizeof(buf))
				{
					fwrite(buf, 1, sizeof(buf), ofp);
				}
				fwrite(buf, 1, size, ofp);
			}
			if (outpath) fclose(ofp);
			fclose(ifp);
			help = 0;
		}
	}
	if (help)
	{
		fprintf(stderr, "Usage: %s [OPTIONS] SOURCE...\n",argv[0]);
		fprintf(stderr, "\tOPTIONS\n");
		fprintf(stderr, "\t\t-o OUTPUT FILENAME\n");
		fprintf(stderr, "\t\t-s SKIP LENGTH IN BYTES\n");
		fprintf(stderr, "\t\t-l OUTPUT LENGTH IN BYTES\n");
		fprintf(stderr, "\t\t-u BLOCK UNIT LENGTH IN BYTES\n");
		fprintf(stderr, "\t\t-n OUTPUT LENGTH IN BLOCKS\n");
		fprintf(stderr, "\t\t-b SKIP LENGTH IN BLOCKS\n");
		fprintf(stderr, "\t\t-p PATCH OFFSET:DATA[,DATA]...\n");
		fprintf(stderr, "\t\t-z ZERO PADDING LENGTH IN BYTES\n");
	}
	return 0;
}
